theory LIA_Checker
  imports Linear_Integer_Arithmetic
    "Simplex/Simplex"
    "~~/src/HOL/Library/Mapping" 
begin

no_notation Simplex.curr_val_satisfies_state ("\<Turnstile>")

subsection {* Entailment checking functionality *}

context LIA begin

subsubsection \<open>Via hints\<close>
fun apply_hint :: "nat list \<Rightarrow> 'v lin_exp list \<Rightarrow> 'v lin_exp" where
  "apply_hint (Cons n ns) (Cons a as) = add_and_mult_list n a (apply_hint ns as)" |
  "apply_hint [] (Cons a as) = add_and_mult_list 1 a (apply_hint [] as)" |
  "apply_hint _ [] = (Lin_Exp 0 [])"

fun lin_exp_is_negative_constant :: "'v :: show lin_exp \<Rightarrow> shows check"
where
"lin_exp_is_negative_constant (Lin_Exp a lst)
 =  do { 
   check (a<0) (shows ''constant remaining must be less than zero'');
   check (Ball (set lst) (const False)) (shows ''coefficients unequal to zero'')
  } <+? (\<lambda> s. shows ''could not derive a contradiction '' o shows_lin_exp (Lin_Exp a lst) o shows_nl o s)"

subsubsection \<open>Via Simplex\<close>
context
  fixes \<rho> :: "'v \<Rightarrow> nat" (* renamings between variables *)
    and \<tau> :: "nat \<Rightarrow> 'v" 
begin
fun lin_list_to_linear_poly :: "('v \<times> int)list \<Rightarrow> linear_poly" where
  "lin_list_to_linear_poly [] = 0" 
| "lin_list_to_linear_poly ((x,c) # le) = lp_monom (of_int c) (\<rho> x) + lin_list_to_linear_poly le" 

fun to_simplex_constraint :: "'v lin_exp \<Rightarrow> constraint" where
  "to_simplex_constraint (Lin_Exp c le) = GEQ (lin_list_to_linear_poly le) (of_int (- c))" 

definition to_rat_assignment :: "('v \<Rightarrow> int) \<Rightarrow> (nat \<Rightarrow> rat)" where
  "to_rat_assignment \<alpha> n = of_int (\<alpha> (\<tau> n))" 

definition lin_exp_vars :: "'v lin_exp \<Rightarrow> 'v set" where
  "lin_exp_vars le = fst ` set (lin_list le)" 

definition lin_exp_vars_list :: "'v lin_exp \<Rightarrow> 'v list" where
  "lin_exp_vars_list le = map fst (lin_list le)" 

lemma lin_exp_vars_list[simp]: "set (lin_exp_vars_list le) = lin_exp_vars le" 
  unfolding lin_exp_vars_def lin_exp_vars_list_def by auto

context
  fixes \<alpha> :: "'v \<Rightarrow> int"
    and \<beta> :: "nat \<Rightarrow> rat"
  assumes beta: "\<beta> = to_rat_assignment \<alpha>" 
begin

lemma lin_list_to_linear_poly: "(\<And> v. v \<in> fst ` set le \<Longrightarrow> \<tau> (\<rho> v) = v) 
  \<Longrightarrow> of_int (interpret_lin_list \<alpha> le) =  ((lin_list_to_linear_poly le) \<lbrace>\<beta>\<rbrace>)" 
proof (induct le)
  case Nil
  show ?case by (simp add: valuate_zero)
next
  case (Cons xc le)
  obtain x c where xc: "xc = (x,c)" by force
  show ?case 
    by (simp add: Cons xc valuate_add valuate_lp_monom beta to_rat_assignment_def)
qed

lemma to_simplex_constraint:
  assumes "interpret_lin_exp_atom \<alpha> le"
    and "(\<And> v. v \<in> lin_exp_vars le \<Longrightarrow> \<tau> (\<rho> v) = v)" 
  shows "\<beta> \<Turnstile>\<^sub>c to_simplex_constraint le" 
proof (cases le)
  case (Lin_Exp x1 x2)
  show ?thesis using assms unfolding Lin_Exp 
    by (auto simp: lin_exp_vars_def lin_list_to_linear_poly[symmetric] interpret_lin_exp_atom_def)
qed

lemma to_simplex_constraint_unsat: 
  assumes "simplex (map to_simplex_constraint \<phi>s) = None" 
    and "\<And> \<phi>. \<phi> \<in> set \<phi>s \<Longrightarrow> interpret_lin_exp_atom \<alpha> \<phi>" 
    and renaming: "(\<And> v \<phi>. v \<in> lin_exp_vars \<phi> \<Longrightarrow> \<phi> \<in> set \<phi>s \<Longrightarrow> \<tau> (\<rho> v) = v)" 
  shows False
proof -
  from to_simplex_constraint[OF assms(2) renaming] 
  have "\<And> \<phi>. \<phi> \<in> set \<phi>s \<Longrightarrow> \<beta> \<Turnstile>\<^sub>c to_simplex_constraint \<phi>" by auto
  hence "\<beta> \<Turnstile>\<^sub>c\<^sub>s map to_simplex_constraint \<phi>s" by simp
  with simplex(2)[OF assms(1)] show False by auto
qed
end
end

definition unsat_via_simplex :: "'v :: show lin_exp list \<Rightarrow> bool" where
  "unsat_via_simplex les = (let vs = remdups (concat (map lin_exp_vars_list les));
     ren_map = Mapping.of_alist (zip vs [0 ..< length vs]);
     ren_fun = (\<lambda> v. case Mapping.lookup ren_map v of None \<Rightarrow> 0 | Some n \<Rightarrow> n);
     cs = map (to_simplex_constraint ren_fun) les
    in simplex cs = None)" 

lemma unsat_via_simplex: assumes unsat: "unsat_via_simplex les" 
  and sat: "\<And> \<phi>. \<phi> \<in> set les \<Longrightarrow> interpret_lin_exp_atom \<alpha> \<phi>"
shows False
proof -
  obtain vs ren_fun inv where
    vs: "vs = remdups (concat (map lin_exp_vars_list les))" and
    ren_fun: "ren_fun = (\<lambda> v. case map_of (zip vs [0 ..< length vs]) v of None \<Rightarrow> 0 | Some n \<Rightarrow> n)" and
    inv: "inv = (\<lambda> i. vs ! i)" by auto  
  have dist: "distinct vs" unfolding vs by auto
  {
    fix v \<phi>
    assume "v \<in> lin_exp_vars \<phi>" "\<phi> \<in> set les" 
    hence "v \<in> set vs" unfolding vs by auto
    then obtain i where i: "i < length vs" and v: "v = vs ! i" unfolding set_conv_nth by auto
    hence "ren_fun (vs ! i) = i" unfolding ren_fun using dist by simp
    hence "inv (ren_fun v) = v" unfolding v inv by simp
  } note inv = this        
  from unsat[unfolded unsat_via_simplex_def Let_def lookup_of_alist, folded vs, folded ren_fun]
  have unsat: "simplex (map (to_simplex_constraint ren_fun) les) = None" by auto
  from to_simplex_constraint_unsat[OF refl unsat sat inv] show False .
qed

end

datatype lia_hints = LIA_Hints "nat list" | LIA_Simplex

instantiation lia_hints :: default begin
  definition "default = LIA_Simplex"
  instance ..
end

derive "show" lia_hints

context LIA begin

fun lia_unsat_checker :: "lia_hints \<Rightarrow> 'v :: {show,linorder} lin_exp list \<Rightarrow> shows check" where
  "lia_unsat_checker hints cnjs = (case hints of 
    LIA_Hints nums_to_multiply \<Rightarrow> or_ok 
    (lin_exp_is_negative_constant (compress (apply_hint nums_to_multiply cnjs)))
    (lin_exp_is_negative_constant (compress (apply_hint (nums_to_multiply @ [1]) cnjs)))
  | LIA_Simplex \<Rightarrow> check (unsat_via_simplex cnjs) (shows ''could not use simplex algorithm to prove unsatisfiability''))
  <+? (\<lambda> s. shows ''The linear inequalities '' \<circ> shows_sep shows_lin_exp shows_nl cnjs \<circ> shows '' cannot be proved unsatisfiable\<newline>'' \<circ>
             shows ''via the hints '' \<circ> shows hints \<circ> shows_nl \<circ> s)" 

(* definitions to avoid the need for existential quantification *)
fun hd2 where "hd2 [] = 1" | "hd2 (Cons a _) = a"

lemma apply_hint_adds:
  "apply_hint h (Cons a as) = add_and_mult_list (hd2 h) a (apply_hint (tl h) as)"
by(cases h,auto)

lemma apply_hint_ok:
  assumes "Ball (set lst) (\<lambda> le. 0 \<le> interpret_lin_exp f le)"
  shows "0 \<le> interpret_lin_exp f (apply_hint h lst)"
using assms proof(induct lst arbitrary:h)
  case (Cons a as)
  hence as_ok: "\<And> h. 0 \<le> interpret_lin_exp f (apply_hint h as)" by auto
  have a_ok: "0 \<le> interpret_lin_exp f a" using Cons.prems by auto
  show ?case unfolding apply_hint_adds using add_and_mult_list_ok[OF a_ok as_ok] by blast
qed(simp)

(* Show that there exists an assignment \<alpha> *)
fun some_alpha::"'a \<times> ty \<Rightarrow> value" where
  "some_alpha (v,IntegerT) = Integer 0" |
  "some_alpha (v,BoolT) = Boolean True"
lemma some_alpha:"assignment some_alpha"
proof
  fix x t show "some_alpha (x::'a, t) \<in> Values_of_type t" by (cases t; simp)
qed

lemma lin_exp_is_negative_constant_unsat:
  assumes "isOK (lin_exp_is_negative_constant cxs)" 
  shows "\<not> interpret_lin_exp_atom \<alpha> cxs"
proof (cases cxs)
  case (Lin_Exp c r)
  have c0:"c < 0" using assms unfolding Lin_Exp by auto
  have "Ball (set r) (const False)" using assms unfolding Lin_Exp by auto
  hence r:"r = []" by fastforce
  have "0 > (interpret_lin_exp \<alpha> (Lin_Exp c r))" by(simp add: r c0)
  thus ?thesis unfolding Lin_Exp interpret_lin_exp_atom_def by auto
qed

lemma entailment_checker_content_checks_for_False:
  assumes "isOK (lin_exp_is_negative_constant (compress (apply_hint h \<phi>s)))"
  and "\<forall> \<phi> \<in> set \<phi>s. interpret_lin_exp_atom \<alpha> \<phi>"
  shows "False"
proof -
  have ths:
    "\<not> (0 \<le> interpret_lin_exp \<alpha> (apply_hint h \<phi>s))"
    using lin_exp_is_negative_constant_unsat[OF assms(1), unfolded interpret_lin_exp_atom_def, unfolded compress_ok] .
  have allHold:
    "Ball (set \<phi>s) (\<lambda> le. 0 \<le> interpret_lin_exp \<alpha> le)"
    using assms(2) unfolding interpret_lin_exp_atom_def by auto
  show ?thesis using ths apply_hint_ok[OF allHold] by auto
qed

lemma lia_unsat_checker: 
  assumes ok: "isOK(lia_unsat_checker h cnjs)"
  and "\<And> c. c \<in> set cnjs \<Longrightarrow> interpret_lin_exp_atom \<alpha> c" 
shows False
proof (cases h)
  case (LIA_Hints hh)
  from ok obtain h2 where isok:"isOK (lin_exp_is_negative_constant
              (compress (apply_hint h2 cnjs)))"
    using assms(1) or_is_or unfolding lia_unsat_checker.simps lia_hints.simps LIA_Hints isOK_update_error by blast
  from entailment_checker_content_checks_for_False[OF this, of \<alpha>] assms(2)
  show False by auto
next
  case LIA_Simplex
  with ok have "unsat_via_simplex cnjs" by auto
  from unsat_via_simplex[OF this assms(2)] show False by auto
qed

subsection \<open>Encoding entailment checking as UNSAT-problems\<close>

fun negate :: "'v::show LIA_exp \<Rightarrow> 'v LIA_exp"
where "negate (Fun Less [a,b]) = Fun Lesseq [b,a]"
    | "negate (Fun Lesseq [a,b]) = Fun Less [b,a]"

fun translate_lit
where "translate_lit (Atom e) = e"
    | "translate_lit (NegAtom e) = negate e"

definition "translate_lits = map translate_lit" 

fun translate_conj
where "translate_conj (Conjunction \<phi>s) = translate_lits \<phi>s"

definition check_lia_clause :: "lia_hints \<Rightarrow> ('v::{show,linorder}) LIA_formula \<Rightarrow> shows check"
where "check_lia_clause hints \<phi> \<equiv> let es = map flatten_exp (translate_conj (\<not>\<^sub>f \<phi>)) in 
    lia_unsat_checker hints es
       <+? (\<lambda> s. shows ''Could not prove unsatisfiability of LIA conjunction '' o shows_nl 
     \<circ> shows_list_gen shows_lin_exp ''False'' [] '' && '' [] es \<circ> shows_nl \<circ> s)"

lemma is_boolE:
  assumes e: "is_bool e"
  and "\<And> s e1 e2. e = Fun s [e1,e2] \<Longrightarrow> s \<in> {Less, Lesseq} \<Longrightarrow> e1 :\<^sub>f IntegerT \<Longrightarrow> e2 :\<^sub>f IntegerT \<Longrightarrow> thesis"
  shows "thesis"
proof -
  from e[unfolded is_bool_def] obtain ty where ty: "e :\<^sub>f ty" "ty \<in> Bool_types" "is_Fun e" by auto
  from ty assms(2) show thesis by (induct e ty rule: has_type_induct, auto)
qed

lemma negate_negates :
  assumes \<alpha>: "assignment \<alpha>" and e: "is_bool e"
  shows "\<alpha> \<Turnstile> Atom (negate e) \<longleftrightarrow> \<not> \<alpha> \<Turnstile> Atom e"
  using assms by (elim is_boolE, auto)

lemma negate_preserves_is_bool:
  assumes e: "is_bool e"
  shows "is_bool (negate e)"
proof -
  from is_boolE [OF e] obtain 
    s e1 e2 where e: "e = Fun s [e1,e2]" and s: "s \<in> { Less, Lesseq}" 
    and e12: "e1 :\<^sub>f IntegerT" "e2 :\<^sub>f IntegerT" .
  then have "i < Suc (Suc 0) \<Longrightarrow> [e2,e1]!i :\<^sub>f [IntegerT,IntegerT]!i" for i by (induct i, auto)
  then show ?thesis unfolding e using s by (auto simp: is_bool_def)
qed

lemma translate_lit:
  assumes "translate_lit \<phi> = e" and "formula \<phi>" and "isLit \<phi>" 
  shows "\<phi> \<Longleftrightarrow>\<^sub>f Atom e" by (cases \<phi>, insert assms, auto dest: negate_negates)

lemma translate_conj:
  assumes es: "translate_conj (\<phi> :: 'v :: show LIA_formula) = es"
      and \<phi>: "formula \<phi>" "is_conj_lits \<phi>" 
  shows "\<phi> \<Longleftrightarrow>\<^sub>f conj_atoms es" "e \<in> set es \<Longrightarrow> is_bool e" 
proof-
  from \<phi>(2,1) obtain ls where \<phi>: "\<phi> = Conjunction ls" and l: "\<And> l. l \<in> set ls \<Longrightarrow> isLit l \<and> formula l" 
    by (cases, auto)
  with es have es: "es = map translate_lit ls" by (auto simp: translate_lits_def)
  {
    fix \<alpha> :: "'v \<times> ty \<Rightarrow> value" assume \<alpha>: "assignment \<alpha>"
    show "\<alpha> \<Turnstile> \<phi> \<longleftrightarrow> \<alpha> \<Turnstile> conj_atoms es"
      using l unfolding \<phi> translate_conj.simps translate_lits_def es
    proof (induct ls)
      case (Cons l ls)
      from Cons(2)[of l] have "formula l" "isLit l" by auto
      from translate_lit[OF refl this \<alpha>] Cons(1)[OF Cons(2)]
      show ?case by auto
    qed auto
  }
  assume "e \<in> set es" 
  thus "is_bool e" using l unfolding \<phi> translate_conj.simps translate_lits_def es
  proof (induct ls)
    case (Cons l ls)
    from Cons(3)[of l] have "formula l" "isLit l" by auto
    hence "is_bool (translate_lit l)" 
      by (cases l, auto intro!: negate_preserves_is_bool)
    with Cons(1)[OF _ Cons(3)] Cons(2) show ?case by auto
  qed auto
qed
  
lemma check_lia_clause:
  assumes ok: "isOK (check_lia_clause h \<phi>)" and \<phi>: "formula \<phi>" and clause: "is_clause \<phi>" and \<alpha>: "assignment \<alpha>"
  shows "satisfies \<alpha> \<phi>"
proof (rule ccontr)
  assume not: "\<not> ?thesis"
  from ok obtain es
  where es: "translate_conj (\<not>\<^sub>f\<phi>) = es"
    and ok: "isOK (lia_unsat_checker h (map flatten_exp es))"
    by (unfold check_lia_clause_def, force)
  from es \<phi> clause have "\<not>\<^sub>f \<phi> \<Longleftrightarrow>\<^sub>f conj_atoms es" by (intro translate_conj, auto)
  from this[OF \<alpha>] not have sat: "\<alpha> \<Turnstile> conj_atoms es" by (auto simp:satisfies_Language)
  show False
  proof(rule lia_unsat_checker[OF ok])
    fix v
    assume v: "v \<in> set (map flatten_exp es)" 
    then obtain w where w: "w \<in> set es" and v: "v = flatten_exp w" by force
    from sat w have eval: "\<alpha> \<Turnstile> Atom w" by auto
    from \<phi> translate_conj(2)[OF es _ _ w] clause
    have "is_bool w" by auto
    from flatten_exp_agrees_bool[OF \<alpha> this] eval
    show "interpret_lin_exp_atom (\<lambda>a. the_int (\<alpha> (a, IntegerT))) v"
      by (auto simp: v interpret_lin_exp_atom_def)
  qed
qed

sublocale logic_checker \<Lambda> check_lia_clause show_LIA_exp by(unfold_locales, fact check_lia_clause)

end
end
