section \<open>Galois Fields\<close>

text \<open>This theory contains definitions of a Galois field ($GF(p^m)$), represented as
  polynomials over GF(p) of degree at most $m$.\<close>

theory Galois_Field
imports
  Polynomial_Field
  Random
begin

typedef GFpn = "UNIV :: GFp poly_f set" by auto 

setup_lifting type_definition_GFpn

instantiation GFpn :: equal
begin
lift_definition equal_GFpn :: "GFpn \<Rightarrow> GFpn \<Rightarrow> bool" is "op =" .
instance
  by (intro_classes, transfer, simp)
end

context
  fixes p pn :: int and F :: "GFp ffield"
  and Fp :: "GFp poly_f"
begin
  
lift_definition plus_GFpn :: "GFpn \<Rightarrow> GFpn \<Rightarrow> GFpn" is
  "plus_poly_f F" .

lift_definition uminus_GFpn :: "GFpn \<Rightarrow> GFpn" is
  "uminus_poly_f F" .

lift_definition minus_GFpn :: "GFpn \<Rightarrow> GFpn \<Rightarrow> GFpn" is
  "minus_poly_f F" .

text \<open>TODO: unsure whether conversion from and to int is okay.\<close>

partial_function (tailrec) of_int_gfpn_main :: "int \<Rightarrow> GFp list \<Rightarrow> GFp list" where
  [code]: "of_int_gfpn_main x ds = (if x = 0 then rev ds else
    let (d,m) = divmod_int x p in 
      of_int_gfpn_main d (of_int_GFp_unsafe m # ds))" 
 
lift_definition of_int_GFpn :: "int \<Rightarrow> GFpn" is
  "\<lambda> x. of_int_gfpn_main (x mod pn) []" .

fun to_int_gfpn_main :: "int \<Rightarrow> GFp list \<Rightarrow> int" where
  "to_int_gfpn_main i (x # xs) = (i * to_int_GFp x + to_int_gfpn_main (i * p) xs)"
| "to_int_gfpn_main _ [] = 0"

lift_definition to_int_GFpn :: "GFpn \<Rightarrow> int" is "to_int_gfpn_main 1" .

lift_definition mult_GFpn :: "GFpn \<Rightarrow> GFpn \<Rightarrow> GFpn" is
  "\<lambda> x y. mod_poly_f F (times_poly_f F x y) Fp" .

lift_definition zero_GFpn :: GFpn is zero_poly_f .
lift_definition one_GFpn :: GFpn is "one_poly_f F" .

definition power_GFpn :: "GFpn \<Rightarrow> nat \<Rightarrow> GFpn" where
  "power_GFpn = fast_power mult_GFpn one_GFpn"

definition inverse_GFpn :: "GFpn \<Rightarrow> GFpn" where
  "inverse_GFpn x = (if x = zero_GFpn then zero_GFpn else power_GFpn x (nat (pn - 2)))"

definition divide_GFpn :: "GFpn \<Rightarrow> GFpn \<Rightarrow> GFpn"  where
  "divide_GFpn x y = mult_GFpn x (inverse_GFpn y)" 

definition GFpn_main :: "GFpn ffield" where
  "GFpn_main = \<lparr> 
    ffield.plus = plus_GFpn,
    minus = minus_GFpn,
    uminus = uminus_GFpn,
    mult = mult_GFpn,
    power = power_GFpn,
    inverse_f = inverse_GFpn,
    divide = divide_GFpn,
    zero = zero_GFpn,
    one = one_GFpn,
    characteristic = p,
    of_int_f = of_int_GFpn,
    to_int_f = to_int_GFpn\<rparr>"
end

context
  fixes F :: "GFp ffield" and one :: GFp and elems :: "GFp list"
begin
private definition generate_candidate :: "GFp list \<Rightarrow> GFp poly_f" where
  "generate_candidate p = rev (one # p)"

partial_function (tailrec) search_irreducible_poly :: "GFp list list \<Rightarrow> nat \<Rightarrow> GFp list \<Rightarrow> 
  bool \<Rightarrow> GFp poly_f" where
  [code]: "search_irreducible_poly left todo p produce = (if produce then
    (if todo = 0 then let pp = generate_candidate p in if irreducible_GFp_poly F pp then pp else 
        search_irreducible_poly left todo p False
     else case elems of
        Nil \<Rightarrow> (* fail *) zero_poly_f
      | y # ys \<Rightarrow> search_irreducible_poly (ys # left) (todo - 1) (y # p) True
    ) else (case left of [] \<Rightarrow> (* fail *) zero_poly_f
    | Cons [] xs \<Rightarrow> search_irreducible_poly xs (Suc todo) (tl p) False
    | Cons (y # ys) xs \<Rightarrow> search_irreducible_poly (ys # xs) todo (y # tl p) True))"
end

fun random_list :: "Random.seed \<Rightarrow> nat \<Rightarrow> nat list \<times> Random.seed" where
  "random_list r 0 = ([],r)"
| "random_list r (Suc n) = (let (x,r') = Random.next r;
     (xs,r'') = random_list r' n
     in (nat_of_natural x # xs, r''))"

partial_function (tailrec) search_irreducible_poly_random 
  :: "GFp ffield \<Rightarrow> Random.seed \<Rightarrow> nat \<Rightarrow> GFp poly_f" where
  [code]: "search_irreducible_poly_random F r n = (let
    (xs,r') = random_list r n;
    p = int_list_to_poly_f F (map int xs @ [1])
    in if irreducible_GFp_poly F p then p else 
    search_irreducible_poly_random F r' n)"    

definition get_irreducible_poly_det :: "GFp ffield \<Rightarrow> nat \<Rightarrow> GFp poly_f" where
  "get_irreducible_poly_det F n = (
    let p = characteristic F; elems = map of_int_GFp_unsafe [0 .. (p-1)]
    in search_irreducible_poly F one_GFp elems [] n [] True)"

definition get_irreducible_poly :: "GFp ffield \<Rightarrow> nat \<Rightarrow> GFp poly_f" where
  "get_irreducible_poly F n = (
    let p = characteristic F; 
      initial_seed = (natural_of_nat 490348901, natural_of_nat 39012)
    in search_irreducible_poly_random F initial_seed n)"

definition GFpn :: "int \<Rightarrow> nat \<Rightarrow> GFpn ffield" where
  "GFpn p n = (let pn = fast_power (op *) 1 p n; F = GFp p;
    Fp = get_irreducible_poly F n
    in GFpn_main p pn F Fp)"

end
